"""
Phenotype Matrix
------------------

A builder that allows the user to set up a
request for a phenotype matrix.

"""

import json
import datetime
import dateutil
import time
import logging
from typing import Callable, Union, Optional, Dict, List
from io import StringIO

from .exceptions import PhenotypeError
from ...exceptions import ServerError
from ...session import ServiceSession
from ...utils import jupyter_available

log = logging.getLogger(__name__)


class PhenotypeMatrix:
    """
    A builder to create a phenotype matrix request.

    You start by using add_phenotype() or add_phenotypes()
    to add a list of phenotypes and then call get_data()
    to retrieve the phenotype matrix from the server.
    """

    def __init__(self, service, base: str = None):
        self.service = service
        self.base = base
        self.phenotypes: Dict[str, Dict[str, Optional[str]]] = {}

    def add_phenotype(
        self,
        name: str,
        missing_value: Optional[str] = None,
        label: Optional[str] = None,
    ) -> None:
        """
        Add a new phenotype to the matrix request.

        :param name: Phenotype name
        :param missing_value: The string to substitute for a missing value in the data
        :param label: Optional label to apply to the phenotype
        """
        self.phenotypes[name] = {
            "name": name,
            "missing_value": missing_value,
            "label": label,
        }

    def add_phenotypes(
        self, names: List[str], missing_value: Optional[str] = None
    ) -> None:
        """
        Add a list of phenotypes to the matrix request.

        :param names: List of phenotype names
        :param missing_value: The string to substitute for a missing value in the data
        """
        for name in names:
            self.add_phenotype(name, missing_value=missing_value)

    def remove_phenotype(self, name: str) -> None:
        """
        Remove a phenotype from the matrix request.

        Does not fail if the phenotype is not present.

        :param name: Phenotype name
        """
        try:
            del self.phenotypes[name]
        except KeyError:
            pass

    def get_data(self, dataframe: bool = True) -> object:
        """
        Retrieve a phenotype matrix from the server.

        Can be called after phenotypes have been added to the request.

        :param dataframe: If set, return results as a pandas dataframe (default True)
        :raises: `PhenotypeError` if the request is not ready or pandas is not installed.
        :raises: `ServerError` if phenotypes are not found on the server
        """
        if not self.phenotypes:
            raise PhenotypeError(
                "Matrix request has not been initialized. Use add_phenotype(s) to begin."
            )
        phenotypes_list = list(self.phenotypes.values())
        content = {
            "base": self.base,
            "phenotypes": phenotypes_list,
        }
        url = self.service.links["get_phenotype_matrix"]
        resp = self.service.session.post(url, json=content)
        resp.raise_for_status()
        tsv_data = resp.text
        if not dataframe:
            return tsv_data

        if not jupyter_available():
            raise PhenotypeError("Pandas library is not installed")

        import pandas as pd

        if not tsv_data:
            return pd.DataFrame()
        df = pd.read_csv(StringIO(tsv_data), delimiter="\t")  # type: ignore
        return df



"""
packagelocal
~~~~~~~~~~~~

Package a local folder and upload to S3 for workflow service and pipelines service.
"""

import os
from time import sleep
import boto3
import socket
from zipfile import ZipFile
import tempfile
import logging
from .exceptions import UploadError
from typing import Dict, Tuple, Sequence, List, Optional, Union, Callable
from .services import BaseService

log = logging.getLogger(__name__)

# !TODO: Temporary until services expose correctly
DEFAULT_SCRATCH_BUCKET = "nextcode-scratch"
EXPIRATION_SECONDS = 7 * 24 * 60 * 60  # expires in 7 days


def package_and_upload(
    service: BaseService, package_name: str, project_path: str
) -> str:
    """
    Create a zip file from a folder and upload to S3. Returns a presigned https URL with an expiration of 24 hours

    :param service: service instance of a workflow service or a pipelines service which supports package uploads
    :param package_name: name of the package, will be used as a partial filename
    :param project_path: local folder to package and upload
    :returns: https url to the zip file on S3
    :raises: UploadError

    """
    log.debug(
        "package_and_upload called with package_name=%s, project_path=%s",
        package_name,
        project_path,
    )
    scratch_bucket = service.app_info.get(
        "scratch_bucket", DEFAULT_SCRATCH_BUCKET
    )  # type: ignore

    if not scratch_bucket:
        raise UploadError(
            "Cannot upload local package because the service has no scratch bucket"
        )

    log.info(
        "Uploading local package %s from %s to s3://%s...",
        package_name,
        project_path,
        scratch_bucket,
    )

    try:
        return _package_and_upload(scratch_bucket, package_name, project_path)
    except Exception as e:
        if "AccessDenied" in repr(
            e
        ) or "The AWS Access Key Id you provided does not exist in our records" in repr(
            e
        ):
            raise UploadError(
                "Failed to upload local package. You do not have access to s3 bucket '%s'."
                % scratch_bucket
            )
        else:
            raise UploadError(f"Failed to upload local package ({repr(e)})")


def _package_and_upload(
    scratch_bucket: str, package_name: str, project_path: str
) -> str:
    """
    Zip up all relevant files from 'project_path' and upload to s3 so that we
    can download it from the ec2 worker node for local deployment.
    """
    log.info("Packaging '%s'", project_path)
    zip_filename = "{}_{}.zip".format(package_name, socket.gethostname())
    full_zip_filename = tempfile.mkstemp(".zip")[1]
    project_path = os.path.abspath(project_path)
    log.debug("project_path is %s", project_path)
    log_archive = ZipFile(full_zip_filename, "w")

    files = []
    for root, directories, filenames in os.walk(
        project_path, followlinks=False, topdown=True
    ):
        # skip everything starting with a .
        if any([l.startswith(".") for l in root.split("/")]):
            continue

        for filename in filenames:
            if not filename.startswith("."):
                full_filename = os.path.join(root, filename)
                files.append(full_filename)

    for f in files:
        write_filename = f.replace(project_path, "")
        write_filename = package_name + write_filename
        log_archive.write(f, arcname=write_filename)
    log_archive.close()
    if len(files) == 0:
        raise RuntimeError("No files found in '%s'" % project_path)
    s3_resource = boto3.resource("s3")
    b = s3_resource.Bucket(scratch_bucket)  # pylint: disable=no-member
    s3_path = "builds/" + zip_filename
    b.upload_file(full_zip_filename, s3_path)

    log.info("Uploaded %s to %s (%s files)", zip_filename, s3_path, len(files))
    s3_client = boto3.client("s3")
    url = s3_client.generate_presigned_url(
        ClientMethod="get_object",
        Params={"Bucket": scratch_bucket, "Key": s3_path},
        ExpiresIn=EXPIRATION_SECONDS,
    )
    # wait for a few seconds to give s3 time to catch up
    sleep(2.0)
    return url



"""
utils
~~~~~~~~~~
Various utilities for nextcode-sdk functionality.
"""

import jwt
import logging
from urllib.parse import urlsplit
import requests
from requests import codes
import binascii
import os
from importlib.util import find_spec

from .exceptions import ServerError, InvalidToken

log = logging.getLogger(__name__)


def decode_token(token):
    try:
        decoded_token = jwt.decode(token, algorithms=["RS256"], verify=False)
        return decoded_token
    except (KeyError, jwt.InvalidTokenError) as ex:
        raise InvalidToken(f"Token could not be decoded ({ex}): {token}")


def check_resp_error(resp):
    response_json = None
    try:
        resp.raise_for_status()
    except Exception:
        desc = resp.text
        try:
            response_json = resp.json()
            desc = response_json["error"]["description"]
            log.info(response_json)
            if "errors" in response_json["error"]:
                desc += " (%s)" % (response_json["error"]["errors"])
        except Exception:
            pass
        if not desc:
            desc = "Status code %s received (%s)" % (resp.status_code, resp.text)
        else:
            desc += " (code %s)" % resp.status_code
        if resp.status_code >= 500:
            desc = "Server error in call to %s" % resp.url
            desc += " - Response headers: %s" % resp.headers
            desc += " - Response body: %s" % resp.text
            log.error(desc)
        else:
            log.info("Server error in call to %s", resp.url)

        error = ServerError(desc, url=resp.url, response=response_json)
        raise error from None


def root_url_from_api_key(api_key):
    payload = decode_token(api_key)
    parts = urlsplit(payload["iss"])
    root_url = "{scheme}://{netloc}".format(scheme=parts.scheme, netloc=parts.netloc)
    return root_url


def get_access_token(api_key):
    """
    """
    payload = decode_token(api_key)
    client_id = payload["azp"]
    body = {
        "grant_type": "refresh_token",
        "client_id": client_id,
        "refresh_token": api_key,
        "username": "dummy_user",
    }
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    token_endpoint = "{}/protocol/openid-connect/token".format(payload["iss"])

    # Call the auth server
    log.info("Authenticating with %s", token_endpoint)
    response = requests.post(token_endpoint, data=body, headers=headers, timeout=5.0)
    if (
        response.status_code == codes.bad_request
        and "Refresh token expired" in response.text
    ):
        raise InvalidToken("Refresh token has expired")
    elif response.status_code >= codes.bad_request:
        try:
            if response.json():
                raise InvalidToken(response.json().get("error_description"))
        except Exception:
            pass

    try:
        response.raise_for_status()
    except Exception:
        log.error("Body: %s" % body)
        raise InvalidToken(
            "Error authenticating with %s: %s" % (token_endpoint, response.text)
        )

    return response.json()["access_token"]


def host_from_url(host):
    """
    Gets the raw URI host of a url with trailing slash, no matter how it is formatted.

    For example, www.server.com/something -> https://www.server.com/
                 http://www.server.com -> http://www.server.com/
                 http://localhost:8080/something -> http://localhost:8080/
    
    """
    if "://" not in host:
        host = f"https://{host}"
    parts = urlsplit(host)
    return "{}://{}/".format(parts.scheme, parts.netloc)


def random_string(num: int = 8) -> str:
    return binascii.hexlify(os.urandom(100)).decode("ascii")[:num]


def jupyter_available() -> bool:
    """
    Check if jupyter dependencies are available without importing these heavy libraries.
    """
    pandas_spec = find_spec("pandas")
    ipython_spec = find_spec("IPython")
    if pandas_spec is not None and ipython_spec is not None:
        return True
    return False



"""
Service class
------------------
Service object for interfacing with the Phenotype Archive API

"""

import os
import logging
from posixpath import join as urljoin
from requests import codes
from typing import Optional, List, Union, Dict

from ...client import Client
from ...services import BaseService
from ...exceptions import ServerError
from .exceptions import PhenotypeError
from .phenotype import Phenotype
from .phenotype_matrix import PhenotypeMatrix

SERVICE_PATH = "api/phenotype-catalog"

log = logging.getLogger(__file__)

SUPPORTED_RESULT_TYPES = ["SET", "QT", "CATEGORY"]


class Service(BaseService):
    """
    A connection to the phenotype catalog service for a specific.

    The project can be passed in e.g. `nextcode.get_service("phenotype", project=myproject)`
    If it is not, the GOR_API_PROJECT environment variable or current profile config will be used.
    A project must be set when the service is instantiated.

    If the project does not already exist in the phenotype catalog only the `create_phenotype`
    method will work and the project will be created implicitly.

    To view available projects use the `svc.all_projects` dict
    """

    project_name: str = ""
    project: Dict = {}
    links: Dict = {}
    all_projects: Dict = {}

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)
        project_name = (
            kwargs.get("project")
            or os.environ.get("GOR_API_PROJECT")
            or client.profile.project
        )
        if not project_name:
            raise PhenotypeError("Please specify a project")

        self._init_project(project_name)

    def _init_project(self, project_name):
        """
        Initialize the project from the server
        """
        self.project_name = project_name
        resp = self.session.get(self.session.url_from_endpoint("projects"))
        self.all_projects = {}
        for project in resp.json()["projects"]:
            self.all_projects[project["name"].lower()] = project

        self.project = self.all_projects.get(self.project_name, {})
        self.links = self.project.get("links", {})
        if self.project:
            log.info(f"Service initialized with project {self.project_name}")
        else:
            log.warning(f"Project {self.project_name} not found")

    def create_phenotype(
        self, name: str, result_type: str, description: Optional[str] = None
    ) -> Phenotype:
        """
        Create a new phenotype in the current project

        :param name: Unique (lowercase) phenotype name in the project
        :raises: PhenotypeError, ServerError
        """
        #! TODO: Get link from root
        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "phenotypes"
        )
        result_type = result_type.upper()
        if result_type not in SUPPORTED_RESULT_TYPES:
            raise PhenotypeError(
                f"Result type {result_type} not supported. Use one of {', '.join(SUPPORTED_RESULT_TYPES)}"
            )
        payload = {"name": name, "result_type": result_type, "description": description}
        resp = self.session.post(url, json=payload)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)
        return Phenotype(self.session, data["phenotype"])

    def get_tags(self) -> List:
        """
        A list of all tags available in the system
        """
        resp = self.session.get(self.session.url_from_endpoint("tags"))
        return resp.json()["tags"]

    def get_phenotypes(self, tags: List[str] = [], limit: int = 100) -> List[Phenotype]:
        """
        A list of all the phenotypes in the current project.

        :param tags: Optional list of tags to filter for
        :param limit: Maximum number of results (default: 100)
        :return: List of phenotypes
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """
        # TODO: project name is case sensitive on the server

        if not self.project:
            raise PhenotypeError("Project does not exist.")
        url = self.links["phenotypes"]
        content = {"with_all_tags": tags, "limit": limit}
        resp = self.session.get(url, data=content)

        data = resp.json()["phenotypes"]
        phenotypes = []
        for item in data:
            phenotypes.append(Phenotype(self.session, item))
        return phenotypes

    def get_phenotype(self, name: str) -> Phenotype:
        """
        Get a specific phenotype in the current project

        :param name: Unique (lowercase) phenotype name in the project
        :return: List of phenotypes as per api spec
        :raises: `PhenotypeError` if the project does not exist
        :raises: `ServerError`
        """
        # TODO: project name is case sensitive on the server
        if not self.project:
            raise PhenotypeError("Project does not exist.")

        url = urljoin(self.links["phenotypes"], name)
        try:
            resp = self.session.get(url)
        except ServerError as ex:
            if ex.response and ex.response["code"] == codes.not_found:
                raise PhenotypeError("Phenotype not found") from None
            else:
                raise

        data = resp.json()["phenotype"]
        return Phenotype(self.session, data)

    def get_phenotype_matrix(self, base: Optional[str] = None) -> PhenotypeMatrix:
        """
        Get a phenotype matrix object.

        :param base: Optional name of base set
        :return: PhenotypeMatrix builder object
        :raises: `PhenotypeError` if the project does not exist
        :raises: `ServerError`
        """
        return PhenotypeMatrix(self, base)



"""
Workflow Job
------------------

The Query class represents a workflow job model from the RESTFul Workflow Service API

"""

import json
import datetime
import dateutil
import time
import logging
from typing import Callable, Union, Optional, Dict, List

from . import RUNNING_STATUSES, FINISHED_STATUSES, FAILED_STATUSES
from .exceptions import JobError
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


class WorkflowJob:
    """
    Proxy object for a serverside workflow job.

    This object can be queried for the current status of a workflow job and will
    automatically refresh from server until the job is finished (or failed)
    """

    def __init__(self, session: ServiceSession, job_id: int, job: Dict):
        self.session = session
        self.job = job
        self.job_id = self.job["job_id"]
        self.links = self.job["links"]

    def __repr__(self) -> str:
        return json.dumps(self.job)

    @property
    def duration(self) -> str:
        """
        Stringified duration of job for human consumption
        """
        if self.complete_date:
            complete_date = self.complete_date
        elif self.status in RUNNING_STATUSES:
            complete_date = datetime.datetime.utcnow()
        elif self.status_date:
            complete_date = self.status_date
        else:
            return "-"
        ret = complete_date - self.submit_date
        # remove microseconds since no one wants them
        ret = ret - datetime.timedelta(microseconds=ret.microseconds)
        return ret

    @property
    def finished(self) -> bool:
        """
        Has the job finished (might be failed)
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FINISHED_STATUSES

    @property
    def running(self) -> bool:
        """
        Is the job currently running
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in RUNNING_STATUSES

    @property
    def failed(self) -> bool:
        """
        Is the job in a failed state
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FAILED_STATUSES

    @property
    def done(self) -> bool:
        """
        Has the job completed successfully
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status == "COMPLETED"

    def refresh(self) -> None:
        """
        Refresh the local cache of the serverside job object
        """
        self.job = self.session.get(self.links["self"]).json()

    def resume(self) -> None:
        """
        Rerun a job that has previously failed
        """
        _ = self.session.put(self.links["self"])
        self.refresh()

    def wait(self, max_seconds: Optional[int] = None, poll_period: float = 0.5):
        """
        Wait for a running job to complete.

        This is similar to the wait method in the Query Service and will wait by default
        indefinitely for a workflow job to complete and poll the server regularly to update
        the local status. When the job is completed (or failed) the method will return the
        workflow job object.

        :param max_seconds: raise an exception if the job runs longer than this
        :param poll_period: Number of seconds to wait between polling (max 10 seconds)
        :returns: WorkflowJob
        :raises: :exc:`JobError`
        """
        if not self.running:
            return self
        log.info("Waiting for job %s to complete...", self.job_id)
        start_time = time.time()
        duration = 0.0
        period = poll_period
        is_running = self.running
        while is_running:
            time.sleep(period)
            duration = time.time() - start_time
            is_running = self.running

            # cancel the wait if the executor pod is in trouble after 30 seconds of waiting to start.
            # it most likely means that the nextflow script has a syntax error or something.
            if self.status == "PENDING" and (
                max_seconds and duration > max_seconds or duration > 30.0
            ):
                log.info("Job has been PENDING for %.0fsec. Inspecting it...", duration)
                curr_status = self.inspect()
                executor_pod = None
                for pod in curr_status["pods"]:
                    if pod["metadata"]["labels"]["app-name"] == "nextflow-executor":
                        executor_pod = pod
                        break
                if not executor_pod:
                    log.warning(
                        "Job is still pending after %.0fsec and executor pod is missing",
                        duration,
                    )
                    raise JobError(
                        f"Job is still pending after {duration:.0f}s and executor pod is missing. View logs or inspect job for failures."
                    )
                if not executor_pod["status"]["container_statuses"][0]["state"][
                    "running"
                ]:
                    log.warning(
                        "Job is still pending after %.0fsec and executor pod is not in running state",
                        duration,
                    )
                    raise JobError(
                        f"Job is still pending after {duration:.0f}s and executor pod is not in running state. View logs or inspect job for failures."
                    )

            if is_running and max_seconds and duration > max_seconds:
                raise JobError(
                    f"Job {self.job_id} has exceeded wait time {max_seconds}s and we will not wait any longer. It is currently {self.status}."
                )

            period = min(period + 0.5, 10.0)
        if self.status == "DONE":
            log.info(
                "Job %s completed in %.2f sec and returned %s rows",
                self.job_id,
                duration,
                self.line_count,
            )
        else:
            log.info(
                "Job %s has status %s after %.2f sec",
                self.job_id,
                self.status,
                duration,
            )
        return self

    def cancel(
        self, status: Optional[str] = None, status_message: Optional[str] = None
    ) -> Optional[str]:
        """
        Cancel a running job

        :param status: status of the cancelled job. Defaults to CANCELLED
        :returns: Status message from the server
        """
        data: Dict = {}
        data["status"] = status or "CANCELLED"
        data["status_message"] = status_message
        resp = self.session.delete(self.links["self"], json=data)
        status_message = resp.json()["status_message"]
        return status_message

    def inspect(self) -> Dict:
        """
        Inspect a failed job for debugging.

        Returns unfiltered pod and node information from the kubernetes system

        :returns: Dictionary containing low-level debugging information
        :raises: :exc:`ServerError` if the server is not configured for inspection capabilities
        """
        try:
            url = self.links["inspect"]
        except KeyError:
            raise ServerError("Server does not support inspect functionality")
        resp = self.session.get(url)
        return resp.json()

    def processes(
        self,
        process_id: Optional[int] = None,
        is_all: bool = False,
        limit: int = 50,
        status: Optional[str] = None,
    ) -> List[Dict]:
        """
        Get a list of all nextflow processes in this job

        :param process_id: process_id to show
        :param is_all: Show all processes, otherwise show only running processes
        :param limit: Maximum number of processes to return
        :param status: Filter processes by status
        """
        url = self.links["processes"]
        if process_id:
            url += "/%s" % process_id
            resp = self.session.get(url)
            return [resp.json()]
        else:
            data: Dict = {"limit": limit}
            if is_all:
                data["all"] = 1
            if status:
                data["status"] = status
            resp = self.session.get(url, json=data)
            return resp.json()["processes"]

    def events(self, limit: int = 50) -> List[Dict]:
        """
        Get a list of events reported by Nextflow for this job

        :param limit: Maximum number of events to return
        """
        url = self.links["events"]
        data = {"limit": limit}
        resp = self.session.get(url, json=data)
        return resp.json()["events"]

    def log_groups(self) -> Dict:
        """
        Get available log groups

        :returns: Dictionary with log group name and server url to the log group
        """
        logs_url = self.links["logs"]
        resp = self.session.get(logs_url)
        return resp.json()["links"]

    def logs(self, log_group: str = "pod", log_filter: Optional[str] = None) -> str:
        """
        Get text logs for the specified log group

        :param log_group: Name of the log group to view
        :param log_filter: Optional filter to apply to the logs
        :raises: :exc:`ServerError` If the log group is not available
        """
        groups = self.log_groups()
        url = None
        for k, v in groups.items():
            if k.startswith(log_group):
                url = v
                break
        if not url:
            raise ServerError(f"Log Group '{log_group}' is not available.")
        if log_filter:
            url += "?filter=%s" % log_filter
        logs = self.session.get(url).text
        return logs

    def __getattr__(self, name):
        try:
            val = self.job[name]
        except KeyError:
            raise AttributeError

        if name.endswith("_date") and val:
            val = dateutil.parser.parse(val)
        return val



"""
csa
~~~~~~~~~~
CSA management features
"""

import os
from urllib.parse import urlsplit
from posixpath import join as urljoin
import requests
from requests import codes
import logging

from .exceptions import ServerError, NotFound, AuthServerError, CSAError
from .utils import host_from_url

log = logging.getLogger(__name__)


def _get_csa_error(resp):
    try:
        msg = resp.json()["error"]["full_message"]
        if isinstance(msg, list):
            msg = ", ".join(msg)
    except Exception:
        msg = str(resp.content)
    return msg


def _check_csa_error(resp):
    if 400 <= resp.status_code < 500:
        raise CSAError(_get_csa_error(resp))
    resp.raise_for_status()


class CSASession:
    def __init__(self, root_url, user_name, password):
        self.root_url = host_from_url(root_url)
        self.session = requests.Session()
        self.session.auth = (user_name, password)
        self.csa_url = urljoin(self.root_url, "csa/api/")

        resp = self.session.get(urljoin(self.csa_url, "users.json"), timeout=2.0)

        if resp.status_code == codes.not_found:
            # ! Temporary hack because services are split between https://[xxx].wuxinextcode.com/
            #   and https://[xxx]-cluster.wuxinextcode.com/
            lst = self.root_url.split(".", 1)
            old_url_base = self.root_url
            if "-cluster" in self.root_url:
                self.root_url = self.root_url.replace("-cluster", "")
            else:
                self.root_url = "{}-cluster.{}".format(lst[0], lst[1])
            log.info(
                "Service not found on server %s. Trying alternative URL %s",
                old_url_base,
                self.root_url,
            )
            self.csa_url = urljoin(self.root_url, "csa/api/")
            resp = self.session.get(urljoin(self.csa_url, "users.json"), timeout=2.0)

        if resp.status_code == codes.unauthorized:
            raise AuthServerError(
                f"User {user_name} could not authenticate with CSA Server"
            )
        resp.raise_for_status()

    def get_user_key(self, user_name):
        users_url = urljoin(self.csa_url, "users.json")
        resp = self.session.get(users_url)

        resp.raise_for_status()
        users = resp.json()["users"]
        for user in users:
            if user["email"] == user_name:
                return user["key"]
        raise AuthServerError(f"User {user_name} not found")

    def create_user(self, user_name, password, exist_ok=False):
        try:
            _ = self.get_user_key(user_name)
        except AuthServerError:
            pass
        else:
            log.info("User '%s' already exists in CSA", user_name)
            if not exist_ok:
                raise CSAError(f"User '{user_name}' already exists.")
            else:
                return

        users_url = urljoin(self.csa_url, "users.json")
        payload = {"user": {"email": user_name, "password": password}}
        resp = self.session.post(users_url, json=payload)
        _check_csa_error(resp)
        log.info("Created user '%s' in CSA", user_name)

    def add_user_to_project(
        self, user_name, project, role="researcher", exist_ok=False
    ):
        user_key = self.get_user_key(user_name)
        roles_url = urljoin(self.csa_url, "user_roles.json")
        resp = self.session.post(
            roles_url,
            json={
                "user_role": {
                    "project_key": project,
                    "role": role,
                    "user_key": user_key,
                }
            },
        )
        if not exist_ok:
            _check_csa_error(resp)
        log.info(
            "User '%s' has been added with role %s to project %s",
            user_name,
            role,
            project,
        )

    def get_project_names(self):
        projects_url = urljoin(self.csa_url, "projects.json")
        resp = self.session.get(projects_url)
        resp.raise_for_status()
        projects = resp.json()["projects"]
        return [p["key"] for p in projects]

    def get_project(self, project_name):
        url = urljoin(self.csa_url, f"projects/{project_name}.json")
        resp = self.session.get(url)
        if resp.status_code == codes.not_found:
            return None
        resp.raise_for_status()
        return resp.json()["project"]

    def create_project(
        self, project_name, org_name, ref_version="hg38", exist_ok=False
    ):
        existing_project = self.get_project(project_name)
        if existing_project:
            if not exist_ok:
                raise CSAError("Project already exists")
            else:
                return existing_project

        url = urljoin(self.csa_url, "projects.json")
        data = {
            "project": {
                "key": project_name,
                "name": project_name,
                "organization_key": org_name,
                "reference_version": ref_version,
            }
        }
        resp = self.session.post(url, json=data)
        _check_csa_error(resp)
        return resp.json()["project"]

    def add_credentials(self, owner_key, service, lookup_key, credential_attributes):
        lookup_key = lookup_key.lower()
        url = urljoin(self.csa_url, "auth/v1/credentials.json")
        url = url.replace("/api", "")

        log.info(
            "Calling '%s' to add '%s' credentials to CSA project '%s'",
            url,
            service,
            owner_key,
        )
        resp = self.session.post(
            url,
            json={
                "credential": {
                    "owner_type": "Project",
                    "owner_key": owner_key,
                    "service": service,
                    "lookup_key": lookup_key,
                    "expires": "",
                    "credential_attributes": credential_attributes,
                }
            },
        )
        _check_csa_error(resp)
        return resp.json()

    def add_s3_credentials(self, owner_key, bucket_name, aws_key, aws_secret):
        return self.add_credentials(
            owner_key, "s3", bucket_name, {"key": aws_key, "secret": aws_secret}
        )

    def add_s3_credentials_to_project(
        self, project_name, bucket_name, aws_key, aws_secret
    ):
        return self.add_s3_credentials(project_name, bucket_name, aws_key, aws_secret)

    def get_s3_credentials_for_project(self, project_name):
        url = urljoin(
            self.csa_url,
            f"auth/v1/credentials.json?find[service]=s3&find[project_key]={project_name}",
        )
        url = url.replace("/api", "")
        resp = self.session.get(url)
        resp.raise_for_status()
        creds = resp.json()["credentials"]
        ret = {}
        for cred in creds:
            ret[cred["lookup_key"]] = {
                "aws_access_key_id": cred["credential_attributes"]["key"],
                "aws_secret_access_key": cred["credential_attributes"]["secret"],
                "project_name": cred["owner_key"],
            }
        return ret

    def import_sample(self, sample_data):
        url = urljoin(self.csa_url, "sample_data_sets.json")
        json_data = {"sample_data_set": sample_data}
        log.info("Importing sample into CSA")
        log.debug("Calling '%s' to import sample: %s", url, json_data)
        resp = self.session.post(url, json=json_data)
        _check_csa_error(resp)
        log.debug("Import sample returned content '%s'", resp.content)
        return resp.json()



"""
config
~~~~~~~~~~
Configuration, caching and profile management for nextcode sdk.

"""

import os
import logging
import yaml
import json
import time
from pathlib import Path
from typing import Dict, Tuple, Sequence, Optional

from .utils import root_url_from_api_key
from .exceptions import InvalidProfile

root_config_folder = Path(os.path.expanduser("~/.nextcode"))

log = logging.getLogger(__name__)

DEFAULT_PROFILE_NAME = "default"
CACHE_SECONDS = 600


def load_cache(name: str) -> Optional[Dict]:
    """
    Load a dictionary from disk cache by name.

    The file is found in ~/.nextcode/cache/[name].cache and is assumed to be a
    dictionary in json format.

    If NEXTCODE_DISABLE_CACHE environment variable is non-zero this method does nothing
    """
    if os.environ.get("NEXTCODE_DISABLE_CACHE"):
        return None
    cache_file = root_config_folder.joinpath("cache", name + ".cache")
    try:
        file_age = time.time() - os.path.getmtime(cache_file)
        if file_age > CACHE_SECONDS:
            log.info("Cache is too old, removing it.")
            os.remove(cache_file)
            raise FileNotFoundError

        contents = json.load(cache_file.open("r"))
        log.info("Loaded contents from cache %s", cache_file)
        return contents
    except FileNotFoundError:
        pass
    except Exception:
        log.exception("Could not load from cache %s", cache_file)
    return None


def save_cache(name: str, contents: Dict) -> None:
    if os.environ.get("NEXTCODE_DISABLE_CACHE"):
        return
    try:
        cache_folder = root_config_folder.joinpath("cache")
        os.makedirs(cache_folder, exist_ok=True)
        cache_file = root_config_folder.joinpath(cache_folder, name + ".cache")
        json.dump(contents, cache_file.open("w"), default=str)
        log.info("Dumped contents into cache %s", cache_file)
    except Exception:
        log.exception("Could not save cache %s", cache_file)


class Config:
    """
    Borg pattern Config class see:
    http://code.activestate.com/recipes/66531-singleton-we-dont-need-no-stinkin-singleton-the-bo/

    Example usage:

    >>> config = Config({'my': 'config'})
    """

    shared_state: Dict = {}
    data: Dict = {}

    def __init__(self, data: Optional[Dict] = None):
        self.__dict__ = self.shared_state
        self.set(data)

    def dict(self) -> Dict:
        return self.data

    def set(self, data: Optional[Dict]) -> None:
        if data is None:
            data = {}
        self.data.update(data)

    def get(self, key: str, default=None):
        return self.data.get(key, default)


def _load_config() -> Dict:
    config_file = root_config_folder.joinpath("config.yaml")
    try:
        content = yaml.safe_load(config_file.open())
        if not isinstance(content, dict):
            raise Exception("Invalid config")
        return content
    except Exception:
        log.info("Config file not found or invalid")
    return {}


def save_config() -> None:
    config = Config()
    config_filename = root_config_folder.joinpath("config.yaml")
    os.makedirs(root_config_folder, exist_ok=True)
    with config_filename.open("w") as conf_file:
        yaml.safe_dump(config.dict(), conf_file)


def _init_config() -> None:
    config = Config()
    config.set({"default_profile": None, "profiles": []})
    content = _load_config()
    if "profiles" not in content:
        content["profiles"] = {}
    for name, profile in content["profiles"].copy().items():
        profile = _prepare_profile(profile)
        if not profile:
            log.warning("Profile '%s' is invalid and will be ignored", name)
            del content["profiles"][name]

    config.set(content)


def _prepare_profile(profile):
    ret = {}
    try:
        ret["api_key"] = profile.get("api_key")
        if profile.get("skip_auth"):
            ret["skip_auth"] = profile.get("skip_auth")
        if profile.get("root_url"):
            ret["root_url"] = profile["root_url"]
        else:
            ret["root_url"] = root_url_from_api_key(ret["api_key"])
    except Exception:
        return None
    return ret


def create_profile(
    name: str, api_key: str, root_url: Optional[str] = None, skip_auth: bool = False
) -> None:
    """
    Create a new profile from api key and persist to disk

    :param name: Unique name of the profile for referencing later
    :param api_key: API Key from keycloak for the server
    :param root_url: root url of the server. If not set, the url from the api key is used
    :param skip_auth: Do not use authentication (local development)
    :raises: InvalidProfile

    """
    contents: Dict = {"api_key": api_key, "root_url": root_url}
    if skip_auth:
        contents["skip_auth"] = True
    profile = _prepare_profile(contents)
    if not profile:
        raise InvalidProfile("Profile does not contain a valid api_key")
    config = Config()
    profiles = config.get("profiles")
    profiles[name] = profile
    config.set({"profiles": profiles})
    save_config()


def delete_profile(name: str) -> None:
    config = Config()
    profiles = get_profiles()
    if name not in profiles:
        raise InvalidProfile("Profile does not exist")
    if config.get("default_profile") == name:
        config.set({"default_profile": None})
    del profiles[name]
    save_config()


def set_default_profile(name: str) -> None:
    """
    Set a named profile as the default one if no profile is specified or GOR_API_KEY is not set

    :param name: Name of the profile
    :raises: InvalidProfile
    """
    config = Config()
    if name not in config.get("profiles"):
        raise InvalidProfile("Profile does not exist")
    config.set({"default_profile": name})
    save_config()


def get_profiles() -> Dict:
    config = Config()
    return config.get("profiles")


def get_default_profile() -> Optional[str]:
    config = Config()
    return os.environ.get("NEXTCODE_PROFILE") or config.get("default_profile")


def get_config() -> Dict:
    config = Config()
    return config.data


def get_profile_config() -> Dict:
    config = Config()
    try:
        return config.data["profiles"][get_default_profile()]
    except KeyError:
        raise InvalidProfile("No current profile found")


_init_config()



"""
Service class
------------------
Service object for interfacing with the GOR Query API.

This class instance is used to communicate with a RESTFul service. Use the `execute` and `execute_template`
methods to get started. To view past queries you can use the `get_queries` and `get_query` methods which return
`Query` objects.

"""

import logging
import os
from typing import Dict, Tuple, Sequence, List, Optional, Union, Any
from requests import codes
from pathlib import Path
import yaml
from urllib.parse import urlencode

from ...services import BaseService
from ...exceptions import ServerError
from ...client import Client
from .exceptions import QueryError, MissingRelations, TemplateError
from .query import Query
from .utils import get_fingerprint
import nextcode

SERVICE_PATH = "api/query"

RUNNING_STATUSES = ("PENDING", "RUNNING", "CANCELLING")
RESULTS_PAGE_SIZE = 200000
QUERY_WAIT_SECONDS = 2
DEFAULT_EXTENSION = ".tsv"

log = logging.getLogger(__name__)


class Service(BaseService):
    """
    A connection to the query api
    """

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)
        self.metadata = {"client": f"nextcode-python-sdk/{nextcode.__version__}"}
        self.project = (
            kwargs.get("project")
            or os.environ.get("GOR_API_PROJECT")
            or client.profile.project
        )

    def _check_project(self):
        """
        Raise an exception if there is no project specified

        :raises: QueryError
        """
        if not self.project:
            raise QueryError("No project specified")

    def set_project(self, project: str, persist: bool = True) -> None:
        """
        Set the current project for all subsequent queries.

        :project: The project name
        :persist: Persist the project name into the current profile

        """
        self.project = project
        if persist:
            self.client.profile.project = self.project

    def get_template(self, name: str) -> Dict:
        """
        Get a specific template from the server based on full name.

        The template name is in the format [organization]/[category]/[name]/[version]
        
        :param name: Full template name
        :raises TemplateError: if the template was not found
        :return: Template dict, see Query API Spec for details
        """
        url = self.session.endpoints["templates"] + name
        try:
            return self.session.get(url).json()
        except ServerError:
            raise TemplateError(f"Template {name} not found")

    def get_templates(
        self, organization: str = None, category: str = None, name: str = None
    ) -> Dict[str, Dict]:
        """
        Return a list of all templates in the system or filtered.

        Returns a list of full template names in the format [organization]/[category]/[name]/[version]

        :param organization: Filter results by organization
        :param category: Filter results by category
        :param name: Filter results by template name
        :raises: ServerError
        """
        if organization:
            links = [self.session.endpoints["templates"] + organization + "/"]
        else:
            orgs = self.session.get(self.session.endpoints["templates"]).json()[
                "organizations"
            ]
            links = [o["links"]["self"] for o in orgs]
        ret = {}
        for link in links:
            if category:
                category_links = [link + category + "/"]
            else:
                category_links = [
                    c["links"]["self"]
                    for c in self.session.get(link).json()["categories"]
                ]
            for link in category_links:
                try:
                    resp = self.session.get(link)
                except ServerError:
                    continue
                for template in resp.json()["templates"]:
                    if not name or template["name"] == name:
                        ret[template["full_name"]] = template
        return ret

    def add_template_from_file(
        self,
        filename: str,
        package_version: Optional[str] = None,
        replace: bool = False,
    ) -> str:
        """
        Add a new template from yaml file.

        If the template already exists a TemplateError is raised. If replace is True
        the template is replaced silently

        :param filename: Full path to a yaml file containing template
        :param package_version: Version of CLA Query Package that contained the Template
        :param replace: Replace an existing template
        :returns: Full name of the new template
        :raises: TemplateError if the file is not found, the template is invalid or if there is a server error.
        """
        p = Path(filename)
        if not p.exists():
            raise TemplateError(f"File '{filename}' not found")
        with p.open() as f:
            yaml_string = f.read()
        try:
            contents = yaml.safe_load(yaml_string)
            if not isinstance(contents, dict):
                raise TemplateError("File contents is not a dictionary")
        except Exception as ex:
            raise TemplateError(f"Yaml is invalid: {ex}")
        name = next(iter(contents))
        url = self.session.url_from_endpoint("templates")
        log.info("Uploading template '%s' to %s...", name, url)
        try:
            resp = self.session.post(
                url, json={"yaml": yaml_string, "package_version": package_version}
            )
        except ServerError as se:
            if se.response and se.response.get("code") == 409:
                if replace:
                    err = se.response["error"]
                    log.info(
                        "Template already exists with ID %s. Replacing...",
                        err["template_id"],
                    )
                    try:
                        resp = self.session.delete(err["template_url"])
                        resp = self.session.post(
                            url,
                            json={
                                "yaml": yaml_string,
                                "package_version": package_version,
                            },
                        )
                    except ServerError as ex:
                        raise TemplateError(str(ex))
                else:
                    raise TemplateError(f"Template '{name}' already exists")
            else:
                raise TemplateError(str(se))

        full_name = resp.json()["full_name"]
        log.info(
            "Template '%s' (%s) has been successfully added",
            full_name,
            resp.json()["id"],
        )
        return full_name

    def delete_template(self, name: str) -> None:
        """
        Deletes a template on the server by full name, e.g. /[org]/[cat]/[name]/[version]

        :param name: Full unique name of the template
        :raises: :exc:`TemplateError`, if the template is not found or if we cannot delete it.
        """
        template = self.get_template(name)
        try:
            _ = self.session.delete(template["links"]["self"])
        except ServerError as e:
            raise TemplateError(f"Could not delete template: {e}")

    def render_template(self, name: str, params: Optional[Dict] = None) -> str:
        """
        Render a template using the supplied arguments and return the full GOR query string.

        :name: Full unique name of the template
        :params: List of arguments to supply to the template
        :raises: :exc:`TemplateError`
        :returns: String containing a fully rendered template
        """
        template = self.get_template(name)

        render_url = template["links"]["render"]
        log.info("Calling render endpoint %s", render_url)
        try:
            resp = self.session.get(
                render_url, params=params, headers={"Accept": "text/plain"}
            )
        except ServerError as ex:
            if "Missing arguments" in str(ex):
                raise TemplateError(str(ex))
            else:
                raise
        return resp.text

    def execute(
        self,
        query: str,
        nowait: bool = False,
        persist: Optional[str] = None,
        relations: Optional[List[Dict]] = None,
        job_type: Optional[str] = None,
        **kw,
    ) -> Query:
        """
        Execute a gor statement on the server.

        :param query: gor query string
        :param nowait: if True, return a Query object with PENDING status instead of waiting for query to finish
        :param relations: virtual relations to include with the query
        :param persist: File path in the project tree to persist the results to (must start with `user_data/`)
        :param job_type: Optional job type for routing purposes

        :raises: :exc:`~exceptions.ServerError`, :exc:`~exceptions.MissingRelations`

        Optional keyword arguments in the form `name=data` are converted into virtual relations if relations
        is not explicitly passed in.

        Whether nowait is set or not, the serverside method will wait for a maximum of 2 seconds for the query
        to transition to a completed status. Therefore, most small queries will return in DONE status.
        """
        self._check_project()
        url = self.session.endpoints["queries"]
        _relations: List[Dict] = []
        if relations:
            _relations = relations
        else:
            for k, v in kw.items():
                _relations.append({"name": f"[{k}]", "data": v})

        payload_relations: List[Dict] = []
        for r in _relations:
            if "data" not in r or "name" not in r:
                raise QueryError("Virtual relations must have name and data fields")
            name = r["name"]
            data = r["data"]

            if hasattr(data, "to_csv"):
                data = data.to_csv(index=False, sep="\t")
            if not isinstance(data, str):
                raise QueryError(f"Virtual relation data for {name} must be a string")

            if not data.startswith("#"):
                data = "#" + data

            fingerprint = r.get("fingerprint") or get_fingerprint(data)
            extension = r.get("extension") or DEFAULT_EXTENSION
            payload_relations.append(
                {
                    "name": name,
                    "fingerprint": fingerprint,
                    "extension": extension,
                    "data": data,
                }
            )
        payload: Dict[str, Optional[Union[int, str, List[Any], Dict]]] = {
            "project": self.project,
            "query": query,
            "relations": payload_relations,
            "persist": persist,
            "wait": QUERY_WAIT_SECONDS,
            "metadata": self.metadata,
            "type": job_type or "default",
        }
        try:
            resp = self.session.post(url, json=payload)
        except ServerError as ex:
            if ex.response and ex.response["code"] == codes.conflict:
                raise MissingRelations(
                    [r["name"] for r in ex.response["error"]["virtual_relations"]]
                )
            else:
                raise
        gor_query = Query(self, resp.json())
        log.info(
            "Query %s has been created and has status %s",
            gor_query.query_id,
            gor_query.status,
        )
        if not nowait:
            gor_query.wait()
        return gor_query

    def execute_template(self, template_name: str, nowait: bool = False, **kw) -> Query:
        """
        Execute a gor template on the server.

        :param template_name: Full template name in the form [org]/[category]/[query]/[version]
        :param nowait: if True, return a Query object with PENDING status instead of waiting for query to finish
        :raises: :exc:`~ServerError`

        Optional keyword arguments are used for arguments into the template.
        """
        self._check_project()
        url = self.session.endpoints["templates"]
        template_url = url + template_name
        try:
            template = self.session.get(template_url).json()
        except ServerError as ex:
            if ex.response and ex.response["code"] == codes.not_found:
                raise QueryError("Template {} not found".format(template_name))
            else:
                raise

        execute_url = template["links"]["execute"]
        args = {}
        for k, v in kw.items():
            args[k] = v
        payload = {
            "project": self.project,
            "args": args,
            "wait": QUERY_WAIT_SECONDS,
            "metadata": self.metadata,
        }
        try:
            resp = self.session.post(execute_url, json=payload)
        except ServerError as ex:
            if "Missing arguments" in repr(ex):
                raise TemplateError(str(ex))
            else:
                raise

        gor_query = Query(self, resp.json())
        log.info(
            "Template Query %s has been created and has status %s",
            gor_query.query_id,
            gor_query.status,
        )
        if not nowait:
            gor_query.wait()
        return gor_query

    def get_query(self, query_id: int) -> Query:
        """
        Get a query object for a query by id, does not require project to be set

        :param query_id: The ID of a query that has been previously executed
        :raises: :exc:`~QueryError`

        """
        url = self.session.endpoints["queries"]
        try:
            resp = self.session.get(f"{url}{query_id}")
        except ServerError as ex:
            if ex.response["code"] == codes.not_found:
                raise QueryError(f"Query {query_id} not found") from None
            raise QueryError(repr(ex), query_id=query_id) from None
        return Query(self, resp.json())

    def get_queries(
        self,
        status: Optional[str] = None,
        user_name: Optional[str] = None,
        limit: int = 100,
        all: Optional[bool] = False,
    ) -> Sequence[Query]:
        """
        Get all queries that have been run by the current user in the current project, optionally filtered by status.

        Results are returned in reverse chronological order so latest queries are first.

        Note that queries are returned in `partial` state which means they might not be up to date. `query.refresh()`
        can be called to force a refresh.

        :param status: Filter queries by status. e.g. `DONE` `RUNNING` `FAILED`.
        :param user_name: Show queries for another user (only available to admin).
        :param limit: Limit the number of queries returned.
        :param all: Fetch all queries in all projects (only available to admin).
        """
        if all:
            user_name = None
            project = None
        else:
            user_name = user_name or self.current_user.get("email")
            project = self.project
        data: Dict[str, Union[str, Any]] = {
            "project": project,
            "user_name": user_name,
            "limit": limit,
            "status": status,
        }
        rsp = self.session.get(self.session.endpoints["queries"], json=data)
        queries = []
        for payload in rsp.json()["queries"]:
            q = Query(self, payload)
            if q.running:
                pass
            queries.append(q)
        return queries



"""
Utility methods
------------------

Utility methods used by the query service.
"""

import hashlib


def get_fingerprint(contents: str) -> str:
    """
    Generate a fingerprint for the contents of a virtual relation.

    This fingerprint is used by the server for caching purposes.

    :param contents: The full contents of a tsv file
    :returns: md5 sum representing the file contents
    """
    md5 = hashlib.md5()
    md5.update(repr(contents).encode())
    return md5.hexdigest()



"""
exceptions
~~~~~~~~~~
Custom exceptions raised by the nextcode-sdk.

"""
from typing import Dict


class ServerError(Exception):
    """
    Raised when there was some sort of an error triggered by the server.
    """

    response: Dict = {}
    url: str = ""

    def __init__(self, message, response=None, url=None, **kw):
        self.response = response
        self.url = url
        self.message = message

    def __str__(self):
        ret = self.message
        if self.url:
            ret += " - {}".format(self.url)
        return ret


class InvalidToken(Exception):
    """
    JWT Token (API Key or access token) is invalid.
    """

    pass


class InvalidProfile(Exception):
    """
    The selected profile is invalid.
    """

    pass


class ServiceNotFound(Exception):
    """
    The requested service is not available on the server.
    """

    pass


class UploadError(Exception):
    """
    The requested server is not available on the server.
    """

    pass


class NotFound(Exception):
    """
    The requested resource was not found
    """

    pass


class AuthServerError(Exception):
    """
    An error with keycloak or CSA management
    """

    pass


class CSAError(Exception):
    """
    An error occurred when interfacing with CSA
    """

    pass



"""
nextcode
~~~~~~~~~~
nextcode is a python module for interfacing with RESTFul services provided by
Wuxi Nextcode.

"""
import os
import logging

from .config import Config

# we want these available from the top-level package
from .client import Client, get_service

log = logging.getLogger()

cfg = Config()


def read_version():
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, "VERSION"), "r") as version_file:
        version = version_file.readline().strip()
        return version


__version__ = read_version()

# loading this here allows easy extension setup in jupyterhub
from .services.query.jupyter import load_ipython_extension
from .services.query import jupyter



"""
Service class
------------------
Service object for interfacing with the Pipelines service API

"""

import logging

from ...client import Client
from ...services import BaseService

RUNNING_STATUSES = ["QUEUED", "LAUNCHING", "INITIALIZING", "RUNNING"]
FAILED_STATUSES = ["FAILED", "TERMINATED"]
FINISHED_STATUSES = ["FINISHED", "CANCELLED"]
ALL_STATUSES = RUNNING_STATUSES + FAILED_STATUSES + FINISHED_STATUSES + ["ALL"]

SERVICE_PATH = "pipelines-service"

log = logging.getLogger(__file__)


class Service(BaseService):
    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)



"""
Exceptions
~~~~~~~~~~

Custom exceptions raised by the pipelines service.

"""

from typing import List


class JobError(Exception):
    def __init__(self, message: str, job_id: int = None):
        self.job_id = job_id



"""
Weblog facilities
------------------

"""
import os
import requests
import datetime
import logging

log = logging.getLogger()


def add_to_details(key, **kw):
    """
    Add a dictionary to a list under 'key' for the job in WEBLOG_URL environment

    :param key: Name of the key to append to
    """
    weblog_url = os.environ.get("WEBLOG_URL")
    if not weblog_url:
        log.warning("No weblog url set. Cannot send message")
        return
    contents = {"event": "custom_details_add", "details": {"key": key, "value": kw}}
    resp = requests.post(weblog_url, json=contents)
    resp.raise_for_status()


def set_details(key, val):
    """
    Set a key under details to a value for the job in WEBLOG_URL environment

    :param key: name of the key in details
    :param val: value to put
    """
    weblog_url = os.environ.get("WEBLOG_URL")
    if not weblog_url:
        log.warning("No weblog url set. Cannot send message")
        return
    contents = {"event": "custom_details_set", "details": {"key": key, "value": val}}
    resp = requests.post(weblog_url, json=contents)
    resp.raise_for_status()


def set_status_message(msg):
    """
    Sets the status message for the job in WEBLOG_URL environment

    :param msg: status message
    """
    weblog_url = os.environ.get("WEBLOG_URL")
    if not weblog_url:
        log.warning("No weblog url set. Cannot send message")
        return
    contents = {"event": "custom_status_message", "details": {"message": msg}}
    resp = requests.post(weblog_url, json=contents)
    resp.raise_for_status()



#!/usr/bin/env python
import os, sys
from setuptools import setup, find_packages, Command

if sys.version_info < (3, 6):
    raise NotImplementedError(
        """Nextcode SDK does not support Python versions older than 3.6"""
    )


version_file = os.path.abspath(os.path.join("nextcode", "VERSION"))


def get_version():
    with open(version_file) as f:
        return f.readlines()[0].strip()


class BumpCommand(Command):
    description = "Version bump"
    user_options = [("which=", None, "Specify which part of the version to bump")]

    def initialize_options(self):
        self.which = None

    def finalize_options(self):
        assert self.which in (None, "major", "minor", "patch"), "Invalid bump pragma"

    def run(self):
        bump_version(self.which)


def bump_version(which):
    major, minor, patch = (0, 1, 0)
    old_version = get_version()
    major, minor, patch = old_version.split(".")

    if which == "major":
        major = int(major) + 1
        minor = 0
        patch = 0
    elif which == "minor":
        minor = int(minor) + 1
        patch = 0
    else:
        patch = int(patch.split("-")[0]) + 1

    new_version = "%s.%s.%s" % (major, minor, patch)

    with open(version_file, "w") as f:
        f.write(new_version)
    print(f"Version has been bumped from {old_version} to {new_version}")


with open("README.md", "r") as fh:
    long_description = fh.read()

version = get_version()

setup(
    name="nextcode-sdk",
    python_requires=">=3.6",
    version=version,
    license="MIT",
    author="WUXI NextCODE",
    author_email="support@wuxinextcode.com",
    description="Python SDK for Wuxi Nextcode Services",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.github.com/wuxi-nextcode/nextcode-python-sdk",
    packages=find_packages(exclude=["tests"]),
    package_data={"nextcode": ["VERSION"]},
    install_requires=[
        "python-dateutil",
        "PyYAML",
        "requests",
        "hjson",
        "PyJWT",
        "boto3",
        "pytest",
    ],
    extras_require={"jupyter": ["pandas", "ipython", "termcolor"]},
    include_package_data=True,
    entry_points={},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    test_suite="tests",
    cmdclass={"bump": BumpCommand},
)



"""
Service class
------------------
Service object for interfacing with the Pipelines Service API.

This class instance is used to communicate with a RESTFul service. The `post_job` method
creates a new job on the server and `find_job` and `get_jobs` allow you to inspect running
and past workflow jobs.

Note: This service has not been fully integrated into the SDK and access is still quite raw. 

"""
import time
import os

from typing import Optional, List, Union, Dict
from ...services import BaseService
from ...client import Client
from ...exceptions import NotFound
from .job import PipelineJob
from .exceptions import JobError
from ...packagelocal import package_and_upload

import logging

SERVICE_PATH = "pipelines-service"


log = logging.getLogger(__name__)


class Service(BaseService):
    """
    A connection to the pipelines service API server
    """

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)

    def get_pipelines(self) -> List:
        """
        Returns the pipelines available on the current server
        
        Refer to the API documentation for the Pipelines service to see formatting of data.

        :return: List of pipelines
        """
        resp = self.session.get(self.session.url_from_endpoint("pipelines"))
        pipelines = resp.json()["pipelines"]
        return pipelines

    def get_projects(self) -> List:
        """
        Returns the projects that have been created on the current server

        Refer to the API documentation for the Pipelines service to see formatting of data.

        :return: List of projects
        """
        resp = self.session.get(self.session.url_from_endpoint("projects"))
        projects = resp.json()["projects"]
        return projects

    def find_job(self, job_id: Union[int, str]) -> PipelineJob:
        """
        Return a job proxy object
        """
        jobs_endpoint = self.session.url_from_endpoint("jobs")

        data: Dict = {"limit": 1}
        if job_id == "latest":
            data["user_name"] = self.current_user.get("email")
        else:
            try:
                data["job_id"] = int(job_id)
            except ValueError:
                raise NotFound(
                    "job_id must be an integer or 'latest', not '%s'" % job_id
                )
        resp = self.session.get(jobs_endpoint, json=data)
        jobs = resp.json()["jobs"]
        if not jobs:
            raise NotFound("Job not found")

        job = jobs[0]
        return PipelineJob(self.session, job["job_id"], job)

    def get_jobs(
        self,
        user_name: Optional[str] = None,
        status: Optional[str] = None,
        project: Optional[str] = None,
        pipeline: Optional[str] = None,
        limit: Optional[int] = 50,
    ) -> List[PipelineJob]:
        """
        Get a list of jobs satisfying the supplied criteria

        :param user_name: The user who created the job
        :param status: Current status of jobs
        :param project: Filter by project
        :param pipeline: Filter by pipeline name
        :param limit: Maximum number of jobs to return
        """
        data: Dict = {"limit": limit}
        if user_name:
            data["user_name"] = user_name
        if status:
            data["status"] = status
        if project:
            data["project_name"] = project
        if pipeline:
            data["pipeline_name"] = pipeline
        st = time.time()
        resp = self.session.get(self.session.url_from_endpoint("jobs"), json=data)
        jobs = resp.json()["jobs"]
        log.info("Retrieved %s jobs in %.2f sec", len(jobs), time.time() - st)
        ret = []
        for job in jobs:
            ret.append(PipelineJob(self.session, job["job_id"], job))
        return ret



"""
Project
------------------

The Project class represents a project model from the RESTFul Project Service API

"""

import json
import datetime
import dateutil
import time
import logging
import os
from typing import Callable, Union, Optional, Dict, List

import boto3
from botocore.client import Config as BotoConfig

from .exceptions import ProjectError
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


class Project:
    """
    Proxy object for a serverside project.
    """

    def __init__(self, service, project: dict):
        self.service = service
        self.session = service.session
        self.project = project
        self.project_name = project["project_name"]

    def __repr__(self) -> str:
        return f"<Project {self.project_name})>"

    def __getattr__(self, name):
        try:
            val = self.project[name]
        except KeyError:
            raise AttributeError

        return val

    def get_users(self):
        # TODO: admin
        users_link = self.links["users"]
        users = self.session.get(users_link)
        return users.json()

    def add_user(self, user_name, policies):
        # TODO: admin
        users_link = self.links["users"]
        data = {"user_name": user_name, "policies": policies}
        try:
            users = self.session.post(users_link, json=data)
        except ServerError as se:
            raise ProjectError(str(se)) from None
        return users.json()

    def remove_user(self, user_name):
        # TODO: admin
        pass

    def delete_project(self):
        # TODO: admin
        pass

    def get_bucket(self):
        credentials = self.service.get_credentials()
        s3 = boto3.resource(
            "s3",
            endpoint_url="https://platform-projects.wuxinextcodedev.com",
            aws_access_key_id=credentials["aws_access_key_id"],
            aws_secret_access_key=credentials["aws_secret_access_key"],
            config=BotoConfig(signature_version="s3v4"),
        )
        bucket = s3.Bucket(self.project_name)  # pylint: disable=E1101
        return bucket

    def list(self, prefix=""):
        bucket = self.get_bucket()
        result = bucket.meta.client.list_objects(
            Bucket=bucket.name, Delimiter="/", Prefix=prefix
        )
        ret = {"folders": [], "files": []}
        if "CommonPrefixes" in result:
            for o in result["CommonPrefixes"]:
                ret["folders"].append(o["Prefix"])
        if "Contents" in result:
            for o in result["Contents"]:
                ret["files"].append(o["Key"])
        return ret

    def download(self, key, path):
        """
        """
        bucket = self.get_bucket()
        path = os.path.expanduser(path)
        if path.endswith("/") or os.path.isdir(path):
            filename = key.split("/")[-1]
            path = os.path.join(path, filename)
        log.info(f"Downloading {key} from project {self.project_name} to {path}")
        bucket.download_file(key, path)

    def upload(self, filename, key):
        bucket = self.get_bucket()
        path = os.path.expanduser(filename)
        if key.endswith("/"):
            key += os.path.basename(filename)
        log.info(f"Uploading {path} to {key} in project {self.project_name}")
        try:
            bucket.upload_file(path, key)
        except Exception as e:
            raise e from None



"""
Service class
------------------
Service object for interfacing with the Project Service API.

"""
import time
import os

from typing import Optional, List, Union, Dict
from ...services import BaseService
from ...client import Client
from ...exceptions import NotFound, ServerError
from .exceptions import ProjectError
from ...utils import jupyter_available

import boto3
from botocore.client import Config as BotoConfig

import logging

SERVICE_PATH = "api/project"
DEFAULT_POLICIES = ["researcher"]

log = logging.getLogger(__name__)


def fmt_size(num, suffix="B"):
    if num == "":
        return ""
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.0f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.0f%s%s" % (num, "Yi", suffix)


class Service(BaseService):
    """
    A connection to the project service API server
    """

    project_name = None
    project = None
    links: dict = {}

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)
        project_name = (
            kwargs.get("project")
            or os.environ.get("GOR_API_PROJECT")
            or client.profile.project
        )

        self.urls = {
            "projects": self.session.url_from_endpoint("projects"),
            "users": self.session.url_from_endpoint("users"),
        }
        self.minio_url = self.session.root_info["app_info"]["minio_url"]
        self.credentials = self.get_credentials(create=True)
        if project_name:
            self._init_project(project_name)

    def _init_project(self, project_name):
        """
        Initialize the project from the server
        """
        self.project_name = project_name
        self.project = self._get_project(project_name)
        self.links = {}
        if self.project:
            self.links = self.project.get("links", {})
            log.info(f"Service initialized with project {self.project_name}")

    def is_admin(self):
        """
        Is the current user a project-service admin
        """
        return self.session.root_info.get("current_user", {}).get("admin", False)

    def create_project(self, project_name: Optional[str] = None):
        """
        Create a new project in the project service.

        :param project_name: The name of the project. If omitted, the current project is assumed
        """
        if not self.is_admin():
            raise ProjectError("You lack the required role to create projects")
        if not project_name:
            project_name = self.project_name
        log.info("Creating project {project_name}")
        return self.session.post(
            self.urls["projects"], json={"project_name": project_name}
        )

    def _check_project(self, check_admin: Optional[bool] = False) -> List:
        """
        Raise an exception if there is no project specified or if the user
        is not a member of the project

        :param check_admin: Only check if the current user is an admin
        :raises: QueryError
        """
        if not self.project:
            raise ProjectError("No project specified")
        if check_admin:
            if self.is_admin():
                return
        self.get_my_project_access()

    def get_user_in_project(self, user_name):
        url = self.links["users"]
        data = {"user_name": user_name}
        resp = self.session.get(url, params=data)
        ret = resp.json()
        if not ret:
            return None
        return ret[0]

    def get_my_project_access(self):
        user = self.get_user_in_project(self.current_user.get("user_name"))
        if not user:
            raise ProjectError(f"You are not a member of project {self.project_name}")
        ret = user["policies"]
        return ret

    def get_all_projects(self) -> List:
        """
        Returns the projects that have been created on the current serv

        :return: List of projects
        """
        resp = self.session.get(self.urls["projects"])
        ret = resp.json()
        return ret

    def _get_project(self, project_name: str) -> Dict:
        data = {"project_name": project_name}
        resp = self.session.get(self.urls["projects"], params=data)
        ret = resp.json()
        if not ret:
            log.warning(
                f"Project {project_name} was not found or you do not have access"
            )
            return ret
        return ret[0]

    def get_my_user(self) -> Dict:
        url = self.current_user["link"]
        resp = self.session.get(url)
        user = resp.json()
        return user

    def get_credentials(self, create: bool = False) -> Dict:
        user = self.get_my_user()
        credentials_url = user["links"]["credentials"]
        try:
            resp = self.session.get(credentials_url)
            credentials = resp.json()
        except ServerError as e:
            if not create:
                raise ProjectError(str(e)) from None
            else:
                credentials = self.set_credentials()
        return credentials

    def set_credentials(self, minio_access_key: Optional[str] = None) -> Dict:
        user = self.get_my_user()
        credentials_url = user["links"]["credentials"]
        try:
            data: Dict = {}
            if minio_access_key:
                data = {"minio_access_key": minio_access_key}

            resp = self.session.put(credentials_url, json=data)
        except ServerError as e:
            raise ProjectError(str(e)) from None

        credentials = resp.json()
        return credentials

    def delete_credentials(self) -> None:
        user = self.get_my_user()
        credentials_url = user["links"]["credentials"]
        _ = self.session.delete(credentials_url)

    def get_users_in_project(self) -> List[Dict]:
        # TODO: admin
        self._check_project(check_admin=True)
        users_link = self.links["users"]
        users = self.session.get(users_link)
        return users.json()

    def add_user_to_project(self, user_name, policies=DEFAULT_POLICIES):
        # TODO: admin
        self._check_project(check_admin=True)
        users_link = self.links["users"]
        data = {"user_name": user_name, "policies": policies}
        try:
            users = self.session.post(users_link, json=data)
        except ServerError as se:
            raise ProjectError(str(se)) from None
        return users.json()

    def remove_user_from_project(self, user_name):
        # TODO: admin
        self._check_project(check_admin=True)
        user = self.get_user_in_project(user_name)
        if not user:
            raise ProjectError(f"User {user_name} not found or you do not have access")
        self.session.delete(user["links"]["self"])

    def obliterate_user(self, user_name):
        # TODO: admin
        url = self.urls["users"]
        data = {"user_name": user_name}
        resp = self.session.get(url, params=data)
        user = resp.json()[0]
        if not user:
            raise ProjectError("User not found")
        self.session.delete(user["links"]["self"])

    def delete_project(self):
        self._check_project()
        # TODO: admin
        raise NotImplementedError("Not yet implemented")

    def get_project_bucket(self):
        self._check_project()
        credentials = self.get_credentials()
        s3 = boto3.resource(
            "s3",
            endpoint_url=self.minio_url,
            aws_access_key_id=credentials["aws_access_key_id"],
            aws_secret_access_key=credentials["aws_secret_access_key"],
            config=BotoConfig(signature_version="s3v4"),
        )  # TODO: Region?
        bucket = s3.Bucket(self.project_name)  # pylint: disable=E1101
        return bucket

    def list(self, prefix: str = "", raw: bool = False):
        self._check_project()
        bucket = self.get_project_bucket()
        result = bucket.meta.client.list_objects(
            Bucket=bucket.name, Delimiter="/", Prefix=prefix
        )
        ret = []
        if "CommonPrefixes" in result:
            for o in result["CommonPrefixes"]:
                ret.append(
                    {"type": "prefix", "size": "", "modified": "", "name": o["Prefix"]}
                )
        if "Contents" in result:
            for o in result["Contents"]:
                ret.append(
                    {
                        "type": "file",
                        "size": o.get("Size"),
                        "modified": o.get("LastModified"),
                        "name": o["Key"],
                    }
                )
        if raw:
            return ret

        if not jupyter_available():
            raise ProjectError(
                "Pandas library is not installed and a dataframe cannot be returned"
            )
        import pandas as pd

        for r in ret:
            r["size"] = fmt_size(r["size"])
            try:
                r["modified"] = r["modified"].strftime("%b %d %H:%M")
            except Exception:
                pass
            if r["type"] == "file":
                r["name"] = r["name"].split("/")[-1]
            else:
                r["name"] = r["name"].split("/")[-2] + "/"
        ret = pd.DataFrame(ret)
        return ret

    def download(self, key, path):
        """
        """
        self._check_project()
        bucket = self.get_project_bucket()
        path = os.path.expanduser(path)
        if path.endswith("/") or os.path.isdir(path):
            filename = key.split("/")[-1]
            path = os.path.join(path, filename)
        log_string = f"Downloading {key} from project {self.project_name} to {path}"
        log.info(log_string)
        bucket.download_file(key, path)

    def upload(self, filename, key):
        self._check_project()
        bucket = self.get_project_bucket()
        path = os.path.expanduser(filename)
        # special case the root folder
        if key in (".", "/"):
            key = ""
        if not key or key.endswith("/"):
            key += os.path.basename(filename)
        if key.startswith("/"):
            key = key[1:]
        # already_exists = not not self.list(key, raw=True)
        # if already_exists and not key.startswith("user_data/"):
        #     raise ProjectError(f"File {key} already exists on server. You can only override files in user_data/")
        log_string = f"Uploading {path} to {key} in project {self.project_name}"
        log.info(log_string)
        try:
            bucket.upload_file(path, key)
        except Exception as e:
            raise e from None

    def delete(self, key):
        pass
